from datetime import datetime, timedelta
from unittest import TestCase
from unittest.mock import MagicMock, patch

from googleapiclient.errors import HttpError

from models.models import YouTubeChannel, OAuthCredentials
from youtube.youtube import (
    pull_youtube_subscriptions,
    get_channels_by_id,
    get_channels_with_new_videos,
    update_channels,
    check_for_new_videos,
    get_most_recent_video,
    calculate_interval_between_cycles,
    _chunk_list,
)


class TestYouTube(TestCase):
    def setUp(self):
        """Set up test fixtures"""
        self.mock_youtube = MagicMock()
        self.sample_channel_id = "UC1234567890"
        self.sample_channel_id_2 = "UC0987654321"

    def tearDown(self):
        """Clean up after tests"""
        pass

    # Tests for pull_youtube_subscriptions
    def test_pull_youtube_subscriptions_single_page(self):
        """Test pulling subscriptions when all results fit in one page"""
        mock_response = {
            "items": [
                {
                    "snippet": {"resourceId": {"channelId": self.sample_channel_id}},
                    "contentDetails": {"totalItemCount": "100"},
                }
            ],
            "pageInfo": {"totalResults": 1},
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        self.mock_youtube.subscriptions().list.return_value = mock_request

        with (
            patch.object(YouTubeChannel, "select") as mock_select,
            patch.object(YouTubeChannel, "create") as mock_create,
        ):
            # Mock the query chain: select().where().exists()
            mock_query = MagicMock()
            mock_query.where.return_value.exists.return_value = False
            mock_select.return_value = mock_query

            # Mock the iteration over select() for the deletion check
            mock_select.return_value.__iter__.return_value = []

            pull_youtube_subscriptions(self.mock_youtube)

            mock_create.assert_called_once_with(
                id=self.sample_channel_id, num_videos=100
            )

    def test_pull_youtube_subscriptions_multiple_pages(self):
        """Test pulling subscriptions with pagination"""
        mock_response_page1 = {
            "items": [
                {
                    "snippet": {"resourceId": {"channelId": self.sample_channel_id}},
                    "contentDetails": {"totalItemCount": "100"},
                }
            ],
            "pageInfo": {"totalResults": 2},
            "nextPageToken": "token123",
        }

        mock_response_page2 = {
            "items": [
                {
                    "snippet": {"resourceId": {"channelId": self.sample_channel_id_2}},
                    "contentDetails": {"totalItemCount": "200"},
                }
            ],
            "pageInfo": {"totalResults": 2},
        }

        mock_request = MagicMock()
        mock_request.execute.side_effect = [mock_response_page1, mock_response_page2]
        self.mock_youtube.subscriptions().list.return_value = mock_request

        with (
            patch.object(YouTubeChannel, "select") as mock_select,
            patch.object(YouTubeChannel, "create") as mock_create,
        ):
            # Mock the query chain: select().where().exists()
            mock_query = MagicMock()
            mock_query.where.return_value.exists.return_value = False
            mock_select.return_value = mock_query

            # Mock the iteration over select() for the deletion check
            mock_select.return_value.__iter__.return_value = []

            pull_youtube_subscriptions(self.mock_youtube)

            assert mock_create.call_count == 2

    def test_pull_youtube_subscriptions_deletes_unsubscribed_channels(self):
        """Test that channels no longer subscribed are deleted"""
        mock_response = {
            "items": [
                {
                    "snippet": {"resourceId": {"channelId": self.sample_channel_id}},
                    "contentDetails": {"totalItemCount": "100"},
                }
            ],
            "pageInfo": {"totalResults": 1},
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        self.mock_youtube.subscriptions().list.return_value = mock_request

        mock_channel = MagicMock()
        mock_channel.id = self.sample_channel_id_2

        with (
            patch.object(YouTubeChannel, "select") as mock_select,
            patch.object(YouTubeChannel, "delete_by_id") as mock_delete,
            patch.object(YouTubeChannel, "create"),
        ):
            # Mock the query chain: select().where().exists()
            mock_query = MagicMock()
            mock_query.where.return_value.exists.return_value = False
            mock_select.return_value = mock_query

            # Mock the iteration over select() for the deletion check
            mock_select.return_value.__iter__.return_value = [mock_channel]

            pull_youtube_subscriptions(self.mock_youtube)

            mock_delete.assert_called_once_with(self.sample_channel_id_2)

    def test_pull_youtube_subscriptions_skips_existing_channels(self):
        """Test that existing channels are not recreated"""
        mock_response = {
            "items": [
                {
                    "snippet": {"resourceId": {"channelId": self.sample_channel_id}},
                    "contentDetails": {"totalItemCount": "100"},
                }
            ],
            "pageInfo": {"totalResults": 1},
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        self.mock_youtube.subscriptions().list.return_value = mock_request

        mock_channel = MagicMock()
        mock_channel.id = self.sample_channel_id

        with (
            patch.object(YouTubeChannel, "select") as mock_select,
            patch.object(YouTubeChannel, "create") as mock_create,
        ):
            # Mock the query chain: select().where().exists()
            mock_query = MagicMock()
            mock_query.where.return_value.exists.return_value = True
            mock_select.return_value = mock_query

            # Mock the iteration over select() for the deletion check
            mock_select.return_value.__iter__.return_value = [mock_channel]

            pull_youtube_subscriptions(self.mock_youtube)

            mock_create.assert_not_called()

    # Tests for get_channels_by_id
    def test_get_channels_by_id_success(self):
        """Test successfully getting channels by ID"""
        mock_response = {
            "items": [
                {
                    "id": self.sample_channel_id,
                    "statistics": {"videoCount": "100"},
                    "snippet": {"title": "Test Channel"},
                }
            ]
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.channels().list.return_value = mock_request

        result = get_channels_by_id(self.mock_youtube, [self.sample_channel_id])

        assert result == mock_response["items"]
        self.mock_youtube.channels().list.assert_called_once_with(
            part="statistics,snippet", id=self.sample_channel_id
        )

    def test_get_channels_by_id_chunks_large_lists(self):
        """Test that channel IDs are chunked when there are more than 50"""
        channel_ids = [f"UC{i:08d}" for i in range(75)]

        mock_response = {"items": [{"id": "test"}]}
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.channels().list.return_value = mock_request

        result = get_channels_by_id(self.mock_youtube, channel_ids)

        # Should be called twice (50 + 25)
        assert self.mock_youtube.channels().list.call_count == 2
        assert len(result) == 2

    def test_get_channels_by_id_no_items_found(self):
        """Test handling when no items are found"""
        mock_response = {}

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.channels().list.return_value = mock_request

        result = get_channels_by_id(self.mock_youtube, [self.sample_channel_id])

        assert result is None

    def test_get_channels_by_id_http_error(self):
        """Test handling of HTTP errors"""
        mock_request = MagicMock()
        mock_request.uri = "https://youtube.com/test"

        http_error = HttpError(resp=MagicMock(status=403), content=b"Quota exceeded")
        mock_request.execute.side_effect = http_error
        self.mock_youtube.channels().list.return_value = mock_request

        result = get_channels_by_id(self.mock_youtube, [self.sample_channel_id])

        assert result is None

    # Tests for get_channels_with_new_videos
    def test_get_channels_with_new_videos_new_video_found(self):
        """Test detecting channels with new videos"""
        mock_prev_channel = MagicMock()
        mock_prev_channel.id = self.sample_channel_id
        mock_prev_channel.num_videos = 100

        current_channels = [
            {
                "id": self.sample_channel_id,
                "statistics": {"videoCount": "101"},
            }
        ]

        result = get_channels_with_new_videos([mock_prev_channel], current_channels)

        assert len(result) == 1
        assert result[0]["id"] == self.sample_channel_id

    def test_get_channels_with_new_videos_no_new_videos(self):
        """Test when no new videos are found"""
        mock_prev_channel = MagicMock()
        mock_prev_channel.id = self.sample_channel_id
        mock_prev_channel.num_videos = 100

        current_channels = [
            {
                "id": self.sample_channel_id,
                "statistics": {"videoCount": "100"},
            }
        ]

        result = get_channels_with_new_videos([mock_prev_channel], current_channels)

        assert len(result) == 0

    def test_get_channels_with_new_videos_video_removed(self):
        """Test handling when videos are removed from a channel"""
        mock_prev_channel = MagicMock()
        mock_prev_channel.id = self.sample_channel_id
        mock_prev_channel.num_videos = 100

        current_channels = [
            {
                "id": self.sample_channel_id,
                "statistics": {"videoCount": "99"},
            }
        ]

        with patch.object(YouTubeChannel, "update") as mock_update:
            mock_update.return_value.where.return_value.execute.return_value = None

            result = get_channels_with_new_videos([mock_prev_channel], current_channels)

            assert len(result) == 0
            mock_update.assert_called_once_with(num_videos=99)

    # Tests for update_channels
    def test_update_channels(self):
        """Test updating channel video counts"""
        channels = [
            {
                "id": self.sample_channel_id,
                "statistics": {"videoCount": "101"},
            }
        ]

        with patch.object(YouTubeChannel, "update") as mock_update:
            mock_update.return_value.where.return_value.execute.return_value = None

            update_channels(channels)

            mock_update.assert_called_once_with(num_videos=101)

    def test_update_channels_multiple(self):
        """Test updating multiple channels"""
        channels = [
            {
                "id": self.sample_channel_id,
                "statistics": {"videoCount": "101"},
            },
            {
                "id": self.sample_channel_id_2,
                "statistics": {"videoCount": "202"},
            },
        ]

        with patch.object(YouTubeChannel, "update") as mock_update:
            mock_update.return_value.where.return_value.execute.return_value = None

            update_channels(channels)

            assert mock_update.call_count == 2

    # Tests for check_for_new_videos
    @patch("youtube.youtube.send_youtube_channels_notifications")
    @patch("youtube.youtube.get_most_recent_video")
    @patch("youtube.youtube.update_channels")
    @patch("youtube.youtube.get_channels_with_new_videos")
    @patch("youtube.youtube.get_channels_by_id")
    def test_check_for_new_videos_single_channel(
        self,
        mock_get_by_id,
        mock_get_new,
        mock_update,
        mock_get_video,
        mock_send_notif,
    ):
        """Test checking for new videos with a single channel"""
        mock_channel = MagicMock()
        mock_channel.id = self.sample_channel_id

        current_channels = [{"id": self.sample_channel_id}]
        new_video_channels = [{"id": self.sample_channel_id}]
        mock_video = {"snippet": {"title": "Test Video"}}

        with patch.object(YouTubeChannel, "select") as mock_select:
            mock_select.return_value = [mock_channel]
            mock_get_by_id.return_value = current_channels
            mock_get_new.return_value = new_video_channels
            mock_get_video.return_value = mock_video

            check_for_new_videos(self.mock_youtube)

            mock_get_video.assert_called_once_with(
                self.mock_youtube, self.sample_channel_id
            )
            mock_send_notif.assert_called_once_with(new_video_channels, mock_video)

    @patch("youtube.youtube.send_youtube_channels_notifications")
    @patch("youtube.youtube.get_most_recent_video")
    @patch("youtube.youtube.update_channels")
    @patch("youtube.youtube.get_channels_with_new_videos")
    @patch("youtube.youtube.get_channels_by_id")
    def test_check_for_new_videos_multiple_channels(
        self,
        mock_get_by_id,
        mock_get_new,
        mock_update,
        mock_get_video,
        mock_send_notif,
    ):
        """Test checking for new videos with multiple channels"""
        mock_channel = MagicMock()
        mock_channel.id = self.sample_channel_id

        current_channels = [
            {"id": self.sample_channel_id},
            {"id": self.sample_channel_id_2},
        ]
        new_video_channels = [
            {"id": self.sample_channel_id},
            {"id": self.sample_channel_id_2},
        ]

        with patch.object(YouTubeChannel, "select") as mock_select:
            mock_select.return_value = [mock_channel]
            mock_get_by_id.return_value = current_channels
            mock_get_new.return_value = new_video_channels

            check_for_new_videos(self.mock_youtube)

            mock_get_video.assert_not_called()
            mock_send_notif.assert_called_once_with(new_video_channels, None)

    @patch("youtube.youtube.send_youtube_channels_notifications")
    @patch("youtube.youtube.update_channels")
    @patch("youtube.youtube.get_channels_with_new_videos")
    @patch("youtube.youtube.get_channels_by_id")
    def test_check_for_new_videos_no_new_videos(
        self, mock_get_by_id, mock_get_new, mock_update, mock_send_notif
    ):
        """Test checking for new videos when there are none"""
        mock_channel = MagicMock()
        mock_channel.id = self.sample_channel_id

        current_channels = [{"id": self.sample_channel_id}]
        new_video_channels = []

        with patch.object(YouTubeChannel, "select") as mock_select:
            mock_select.return_value = [mock_channel]
            mock_get_by_id.return_value = current_channels
            mock_get_new.return_value = new_video_channels

            check_for_new_videos(self.mock_youtube)

            mock_send_notif.assert_not_called()

    # Tests for get_most_recent_video
    def test_get_most_recent_video_success(self):
        """Test successfully getting the most recent video"""
        now = datetime.now()
        published_at = (now - timedelta(minutes=1)).strftime("%Y-%m-%dT%H:%M:%SZ")

        mock_response = {
            "items": [
                {
                    "snippet": {
                        "publishedAt": published_at,
                        "position": 0,
                    },
                    "status": {"privacyStatus": "public"},
                }
            ]
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.playlistItems().list.return_value = mock_request

        result = get_most_recent_video(self.mock_youtube, self.sample_channel_id)

        assert result == mock_response["items"][0]

    def test_get_most_recent_video_not_public(self):
        """Test when the most recent video is not public"""
        now = datetime.now()
        published_at = (now - timedelta(minutes=1)).strftime("%Y-%m-%dT%H:%M:%SZ")

        mock_response = {
            "items": [
                {
                    "snippet": {
                        "publishedAt": published_at,
                        "position": 0,
                    },
                    "status": {"privacyStatus": "private"},
                }
            ]
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.playlistItems().list.return_value = mock_request

        result = get_most_recent_video(self.mock_youtube, self.sample_channel_id)

        assert result is None

    def test_get_most_recent_video_too_old(self):
        """Test when the most recent video is too old"""
        now = datetime.now()
        published_at = (now - timedelta(minutes=5)).strftime("%Y-%m-%dT%H:%M:%SZ")

        mock_response = {
            "items": [
                {
                    "snippet": {
                        "publishedAt": published_at,
                        "position": 0,
                    },
                    "status": {"privacyStatus": "public"},
                }
            ]
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.playlistItems().list.return_value = mock_request

        result = get_most_recent_video(self.mock_youtube, self.sample_channel_id)

        assert result is None

    def test_get_most_recent_video_wrong_position(self):
        """Test when the video is not in position 0"""
        now = datetime.now()
        published_at = (now - timedelta(minutes=1)).strftime("%Y-%m-%dT%H:%M:%SZ")

        mock_response = {
            "items": [
                {
                    "snippet": {
                        "publishedAt": published_at,
                        "position": 1,
                    },
                    "status": {"privacyStatus": "public"},
                }
            ]
        }

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.playlistItems().list.return_value = mock_request

        result = get_most_recent_video(self.mock_youtube, self.sample_channel_id)

        assert result is None

    def test_get_most_recent_video_no_items(self):
        """Test when no items are returned"""
        mock_response = {}

        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_request.uri = "https://youtube.com/test"
        self.mock_youtube.playlistItems().list.return_value = mock_request

        result = get_most_recent_video(self.mock_youtube, self.sample_channel_id)

        assert result is None

    def test_get_most_recent_video_http_error(self):
        """Test handling of HTTP errors"""
        mock_request = MagicMock()
        mock_request.uri = "https://youtube.com/test"

        http_error = HttpError(
            resp=MagicMock(status=404), content=b"Playlist not found"
        )
        mock_request.execute.side_effect = http_error
        self.mock_youtube.playlistItems().list.return_value = mock_request

        result = get_most_recent_video(self.mock_youtube, self.sample_channel_id)

        assert result is None

    # Tests for calculate_interval_between_cycles
    def test_calculate_interval_between_cycles_single_key(self):
        """Test interval calculation with a single API key"""
        mock_channel = MagicMock()
        mock_channel.id = self.sample_channel_id
        mock_cred = MagicMock()

        with (
            patch.object(YouTubeChannel, "select") as mock_channel_select,
            patch.object(OAuthCredentials, "select") as mock_cred_select,
        ):
            mock_channel_select.return_value = [mock_channel]
            mock_cred_select.return_value = [mock_cred]

            interval = calculate_interval_between_cycles()

            # With 1 channel and 1 API key:
            # requests_per_cycle = 1 (rounded up from (1+1)/50)
            # num_cycles_per_day = 10000 / 1 = 10000
            # interval = 86400 / 10000 = 8.64 seconds
            assert interval == 9  # Rounded up

    def test_calculate_interval_between_cycles_multiple_keys(self):
        """Test interval calculation with multiple API keys"""
        mock_channels = [MagicMock() for _ in range(100)]
        mock_creds = [MagicMock() for _ in range(2)]

        with (
            patch.object(YouTubeChannel, "select") as mock_channel_select,
            patch.object(OAuthCredentials, "select") as mock_cred_select,
        ):
            mock_channel_select.return_value = mock_channels
            mock_cred_select.return_value = mock_creds

            interval = calculate_interval_between_cycles()

            # With 100 channels and 2 API keys:
            # requests_per_cycle = 3 (rounded up from (100+1)/50)
            # total_requests = 2 * 10000 = 20000
            # num_cycles_per_day = 20000 / 3 = 6666
            # interval = 86400 / 6666 = 12.96 seconds
            assert interval == 13  # Rounded up

    def test_calculate_interval_between_cycles_many_channels(self):
        """Test interval calculation with many channels"""
        mock_channels = [MagicMock() for _ in range(500)]
        mock_creds = [MagicMock()]

        with (
            patch.object(YouTubeChannel, "select") as mock_channel_select,
            patch.object(OAuthCredentials, "select") as mock_cred_select,
        ):
            mock_channel_select.return_value = mock_channels
            mock_cred_select.return_value = mock_creds

            interval = calculate_interval_between_cycles()

            # With 500 channels and 1 API key:
            # requests_per_cycle = 11 (rounded up from (500+1)/50)
            # num_cycles_per_day = 10000 / 11 = 909
            # interval = 86400 / 909 = 95.05 seconds
            assert interval == 96  # Rounded up

    # Tests for _chunk_list
    def test_chunk_list_less_than_chunk_size(self):
        """Test chunking a list smaller than chunk size"""
        test_list = ["id1", "id2", "id3"]
        result = list(_chunk_list(test_list))

        assert len(result) == 1
        assert result[0] == "id1,id2,id3"

    def test_chunk_list_exactly_chunk_size(self):
        """Test chunking a list exactly equal to chunk size"""
        test_list = [f"id{i}" for i in range(50)]
        result = list(_chunk_list(test_list))

        assert len(result) == 1
        assert len(result[0].split(",")) == 50

    def test_chunk_list_multiple_chunks(self):
        """Test chunking a list into multiple chunks"""
        test_list = [f"id{i}" for i in range(125)]
        result = list(_chunk_list(test_list))

        assert len(result) == 3
        assert len(result[0].split(",")) == 50
        assert len(result[1].split(",")) == 50
        assert len(result[2].split(",")) == 25

    def test_chunk_list_custom_chunk_size(self):
        """Test chunking with a custom chunk size"""
        test_list = [f"id{i}" for i in range(30)]
        result = list(_chunk_list(test_list, chunk_size=10))

        assert len(result) == 3
        assert len(result[0].split(",")) == 10
        assert len(result[1].split(",")) == 10
        assert len(result[2].split(",")) == 10

    def test_chunk_list_empty_list(self):
        """Test chunking an empty list"""
        test_list = []
        result = list(_chunk_list(test_list))

        assert len(result) == 0



from unittest import TestCase
from unittest.mock import MagicMock, patch

from util.healthcheck import healthcheck


class TestHealthcheck(TestCase):
    def setUp(self):
        """Set up test fixtures"""
        self.example_channel_id = "UC_x5XG1OV2P6uZZ5FSM9Ttw"
        self.mock_youtube = MagicMock()

    def tearDown(self):
        """Clean up after tests"""
        pass

    # Tests for healthcheck
    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_success(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test successful healthcheck"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service
        mock_youtube = MagicMock()
        mock_oauth.get_authenticated_youtube_service.return_value = mock_youtube

        # Mock YouTube API response
        mock_response = {
            "items": [{"id": self.example_channel_id}],
            "pageInfo": {"totalResults": 1},
        }
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_youtube.channels().list.return_value = mock_request

        # Test healthcheck and expect exit(0)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 0)
        mock_logger.info.assert_called_once_with("Healthcheck passed.")

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.logger")
    def test_healthcheck_youtube_channel_table_missing(
        self, mock_logger, mock_database, mock_oauth
    ):
        """Test healthcheck fails when YouTubeChannel table doesn't exist"""
        # Mock database table existence - YouTubeChannel missing
        mock_database.table_exists.side_effect = lambda table: False

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("YouTubeChannel table does not exist", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.logger")
    def test_healthcheck_oauth_credentials_table_missing(
        self, mock_logger, mock_database, mock_oauth
    ):
        """Test healthcheck fails when oauth_credentials table doesn't exist"""

        # Mock database table existence - oauth_credentials missing
        def table_exists_side_effect(table):
            if table == "youtubechannel":
                return True
            return False

        mock_database.table_exists.side_effect = table_exists_side_effect

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("OAuthCredentials table does not exist", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_no_oauth_credentials(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test healthcheck fails when no OAuth credentials exist"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials - empty
        mock_oauth_creds.select.return_value = []

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("No OAuth credentials found in database", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_no_youtube_service(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test healthcheck fails when YouTube service is not available"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service - None
        mock_oauth.get_authenticated_youtube_service.return_value = None

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("No valid YouTube service available", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_channel_not_found_no_items(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test healthcheck fails when channel is not found (no items in response)"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service
        mock_youtube = MagicMock()
        mock_oauth.get_authenticated_youtube_service.return_value = mock_youtube

        # Mock YouTube API response - no items
        mock_response = {}
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_youtube.channels().list.return_value = mock_request

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("Healthcheck channel not found", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_channel_not_found_empty_items(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test healthcheck fails when channel is not found (empty items list)"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service
        mock_youtube = MagicMock()
        mock_oauth.get_authenticated_youtube_service.return_value = mock_youtube

        # Mock YouTube API response - empty items
        mock_response = {"items": [], "pageInfo": {"totalResults": 0}}
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_youtube.channels().list.return_value = mock_request

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("Healthcheck channel not found", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_channel_not_found_zero_results(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test healthcheck fails when channel is not found (zero total results)"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service
        mock_youtube = MagicMock()
        mock_oauth.get_authenticated_youtube_service.return_value = mock_youtube

        # Mock YouTube API response - zero total results
        mock_response = {
            "items": [{"id": self.example_channel_id}],
            "pageInfo": {"totalResults": 0},
        }
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_youtube.channels().list.return_value = mock_request

        # Test healthcheck and expect exit(1)
        with self.assertRaises(SystemExit) as cm:
            healthcheck()

        self.assertEqual(cm.exception.code, 1)
        mock_logger.error.assert_called_once()
        error_msg = mock_logger.error.call_args[0][0]
        self.assertIn("Healthcheck channel not found", error_msg)

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_verifies_correct_channel_id(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test that healthcheck calls YouTube API with correct channel ID"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service
        mock_youtube = MagicMock()
        mock_oauth.get_authenticated_youtube_service.return_value = mock_youtube

        # Mock YouTube API response
        mock_response = {
            "items": [{"id": self.example_channel_id}],
            "pageInfo": {"totalResults": 1},
        }
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_youtube.channels().list.return_value = mock_request

        # Test healthcheck
        with self.assertRaises(SystemExit):
            healthcheck()

        # Verify YouTube API was called with correct parameters
        mock_youtube.channels().list.assert_called_once_with(
            part="id", id=self.example_channel_id
        )

    @patch("util.healthcheck.oauth")
    @patch("util.healthcheck.database")
    @patch("util.healthcheck.OAuthCredentials")
    @patch("util.healthcheck.logger")
    def test_healthcheck_calls_oauth_with_force_auth(
        self, mock_logger, mock_oauth_creds, mock_database, mock_oauth
    ):
        """Test that healthcheck calls OAuth with force_auth=True"""
        # Mock database table existence
        mock_database.table_exists.side_effect = lambda table: True

        # Mock OAuth credentials
        mock_creds = MagicMock()
        mock_oauth_creds.select.return_value = [mock_creds]

        # Mock YouTube service
        mock_youtube = MagicMock()
        mock_oauth.get_authenticated_youtube_service.return_value = mock_youtube

        # Mock YouTube API response
        mock_response = {
            "items": [{"id": self.example_channel_id}],
            "pageInfo": {"totalResults": 1},
        }
        mock_request = MagicMock()
        mock_request.execute.return_value = mock_response
        mock_youtube.channels().list.return_value = mock_request

        # Test healthcheck
        with self.assertRaises(SystemExit):
            healthcheck()

        # Verify OAuth was called with force_auth=True
        mock_oauth.get_authenticated_youtube_service.assert_called_once_with(
            force_auth=True
        )



import sys

from dotenv import load_dotenv

from auth import oauth
from util.healthcheck import healthcheck
import time

from models import database
from models.models import YouTubeChannel, OAuthCredentials
from util.logging import logger
from youtube.youtube import (
    check_for_new_videos,
    calculate_interval_between_cycles,
    pull_youtube_subscriptions,
)


def main():
    logger.info("Staring BSN...")
    if not database.table_exists("youtubechannel"):
        logger.info("YouTube Channels table does not exist. Creating table...")
        database.create_tables([YouTubeChannel])
        database.commit()
    if not database.table_exists("oauth_credentials"):
        logger.info("OAuth Credentials table does not exist. Creating table...")
        database.create_tables([OAuthCredentials])
        database.commit()
        oauth.get_authenticated_youtube_service(force_auth=True)

    interval_between_checks: int = calculate_interval_between_cycles()

    while True:
        youtube = oauth.get_authenticated_youtube_service(force_auth=True)
        if youtube:
            pull_youtube_subscriptions(youtube)
            # TODO More work could be done to reduce API calls, by using `subscription` to check video counts
            check_for_new_videos(youtube)
            logger.info(f"Sleeping for {interval_between_checks} seconds...")
            time.sleep(interval_between_checks)
        else:
            logger.error("No valid credentials available. Exiting.")
            exit(1)


if __name__ == "__main__":
    load_dotenv()
    if len(sys.argv) > 1 and sys.argv[1] == "healthcheck":
        healthcheck()
    else:
        main()



from datetime import datetime, timezone, timedelta
from unittest import TestCase
from unittest.mock import MagicMock, patch

import requests
from google.auth.credentials import TokenState
from google.oauth2.credentials import Credentials

from auth.oauth import (
    get_authenticated_youtube_service,
    run_auth_flow,
    save_credentials_to_db,
    load_credentials_from_db,
    get_credentials,
    revoke_credentials,
    _now_utc,
    DEVICE_CODE_URL,
    TOKEN_URL,
    REVOKE_URL,
    DEFAULT_SCOPES,
)


class TestOAuth(TestCase):
    def setUp(self):
        """Set up test fixtures"""
        self.client_id = "test_client_id.apps.googleusercontent.com"
        self.client_secret = "test_client_secret"
        self.user_id = "test_user_id"
        self.user_email = "test@example.com"
        self.access_token = "test_access_token"
        self.refresh_token = "test_refresh_token"
        self.scopes = ["https://www.googleapis.com/auth/youtube.readonly"]

        # Create a mock credentials object
        self.mock_creds = MagicMock(spec=Credentials)
        self.mock_creds.token = self.access_token
        self.mock_creds.refresh_token = self.refresh_token
        self.mock_creds.token_uri = TOKEN_URL
        self.mock_creds.client_id = self.client_id
        self.mock_creds.client_secret = self.client_secret
        self.mock_creds.scopes = self.scopes
        self.mock_creds.expiry = datetime.now(timezone.utc) + timedelta(hours=1)
        self.mock_creds.token_state = TokenState.FRESH

    def tearDown(self):
        """Clean up after tests"""
        pass

    # Tests for _now_utc
    def test_now_utc_returns_aware_datetime(self):
        """Test that _now_utc returns a timezone-aware datetime"""
        now = _now_utc()
        self.assertIsInstance(now, datetime)
        self.assertIsNotNone(now.tzinfo)
        self.assertEqual(now.tzinfo, timezone.utc)

    def test_now_utc_returns_current_time(self):
        """Test that _now_utc returns approximately the current time"""
        before = datetime.now(timezone.utc)
        result = _now_utc()
        after = datetime.now(timezone.utc)

        self.assertGreaterEqual(result, before)
        self.assertLessEqual(result, after)

    # Tests for get_authenticated_youtube_service
    @patch("auth.oauth.get_credentials")
    @patch("auth.oauth.build")
    def test_get_authenticated_youtube_service_with_credentials(
        self, mock_build, mock_get_creds
    ):
        """Test getting YouTube service with existing credentials"""
        mock_get_creds.return_value = self.mock_creds
        mock_service = MagicMock()
        mock_build.return_value = mock_service

        result = get_authenticated_youtube_service()

        self.assertEqual(result, mock_service)
        mock_get_creds.assert_called_once()
        mock_build.assert_called_once_with("youtube", "v3", credentials=self.mock_creds)

    @patch("auth.oauth.get_credentials")
    def test_get_authenticated_youtube_service_no_credentials_no_force(
        self, mock_get_creds
    ):
        """Test getting YouTube service without credentials and force_auth=False"""
        mock_get_creds.return_value = None

        result = get_authenticated_youtube_service(force_auth=False)

        self.assertIsNone(result)
        mock_get_creds.assert_called_once()

    @patch("auth.oauth.run_auth_flow")
    @patch("auth.oauth.get_credentials")
    @patch("auth.oauth.build")
    def test_get_authenticated_youtube_service_force_auth(
        self, mock_build, mock_get_creds, mock_run_auth
    ):
        """Test getting YouTube service with force_auth=True"""
        mock_get_creds.return_value = None
        mock_run_auth.return_value = self.mock_creds
        mock_service = MagicMock()
        mock_build.return_value = mock_service

        result = get_authenticated_youtube_service(
            force_auth=True,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

        self.assertEqual(result, mock_service)
        mock_run_auth.assert_called_once_with(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=list(DEFAULT_SCOPES),
        )

    @patch("auth.oauth.get_credentials")
    @patch("auth.oauth.build")
    def test_get_authenticated_youtube_service_build_fails(
        self, mock_build, mock_get_creds
    ):
        """Test when building YouTube service fails"""
        mock_get_creds.return_value = self.mock_creds
        mock_build.side_effect = Exception("Build failed")

        result = get_authenticated_youtube_service()

        self.assertIsNone(result)

    @patch("auth.oauth.get_credentials")
    @patch.dict(
        "os.environ",
        {"OAUTH_CLIENT_ID": "env_client_id", "OAUTH_CLIENT_SECRET": "env_secret"},
    )
    def test_get_authenticated_youtube_service_uses_env_vars(self, mock_get_creds):
        """Test that service uses environment variables when params not provided"""
        mock_get_creds.return_value = None

        result = get_authenticated_youtube_service(force_auth=False)

        # Should log about checking environment variables
        self.assertIsNone(result)

    @patch("auth.oauth.run_auth_flow")
    @patch("auth.oauth.get_credentials")
    def test_get_authenticated_youtube_service_custom_scopes(
        self, mock_get_creds, mock_run_auth
    ):
        """Test getting YouTube service with custom scopes"""
        mock_get_creds.return_value = None
        custom_scopes = ["https://www.googleapis.com/auth/youtube"]
        mock_run_auth.return_value = self.mock_creds

        get_authenticated_youtube_service(
            force_auth=True,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=custom_scopes,
        )

        mock_run_auth.assert_called_once_with(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=custom_scopes,
        )

    # Tests for run_auth_flow
    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.save_credentials_to_db")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.time.sleep")
    def test_run_auth_flow_success(
        self, mock_sleep, mock_load_creds, mock_save_creds, mock_post
    ):
        """Test successful OAuth device flow"""
        # Mock device code response
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        # Mock token response (success on first try)
        token_response = MagicMock()
        token_response.status_code = 200
        token_response.json.return_value = {
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "expires_in": 3600,
        }

        mock_post.side_effect = [device_response, token_response]
        mock_load_creds.return_value = self.mock_creds

        result = run_auth_flow(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=self.scopes,
        )

        self.assertEqual(result, self.mock_creds)
        self.assertEqual(mock_post.call_count, 2)

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.OAuthCredentials")
    def test_run_auth_flow_no_client_id(self, mock_oauth_creds, mock_post):
        """Test run_auth_flow raises error when no client_id available"""
        mock_oauth_creds.select().first.return_value = None

        with self.assertRaises(ValueError) as context:
            run_auth_flow(client_id=None)

        self.assertIn("client_id must be provided", str(context.exception))

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.OAuthCredentials")
    def test_run_auth_flow_uses_db_client_id(self, mock_oauth_creds, mock_post):
        """Test run_auth_flow uses client_id from database when not provided"""
        mock_row = MagicMock()
        mock_row.client_id = self.client_id
        mock_oauth_creds.select().first.return_value = mock_row

        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()
        mock_post.return_value = device_response

        with patch("auth.oauth.time.sleep"):
            with patch("auth.oauth.time.time", side_effect=[0, 1000]):
                # Will timeout
                run_auth_flow(client_id=None)

        # Should have attempted with DB client_id
        mock_post.assert_called()

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.sleep")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_timeout(self, mock_time, mock_sleep, mock_post):
        """Test run_auth_flow times out waiting for authorization"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        token_response = MagicMock()
        token_response.status_code = 400
        token_response.json.return_value = {"error": "authorization_pending"}

        mock_post.side_effect = [device_response] + [token_response] * 10
        mock_time.side_effect = [0, 0] + [1000] * 20  # Simulate timeout

        result = run_auth_flow(
            client_id=self.client_id, client_secret=self.client_secret, timeout=10
        )

        self.assertIsNone(result)

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.save_credentials_to_db")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.time.sleep")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_authorization_pending(
        self, mock_time, mock_sleep, mock_load_creds, mock_save_creds, mock_post
    ):
        """Test run_auth_flow handles authorization_pending correctly"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        pending_response = MagicMock()
        pending_response.status_code = 400
        pending_response.json.return_value = {"error": "authorization_pending"}

        success_response = MagicMock()
        success_response.status_code = 200
        success_response.json.return_value = {
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "expires_in": 3600,
        }

        mock_post.side_effect = [device_response, pending_response, success_response]
        mock_load_creds.return_value = self.mock_creds
        mock_time.side_effect = [0, 1, 2, 3, 4, 5]

        result = run_auth_flow(client_id=self.client_id)

        self.assertEqual(result, self.mock_creds)
        mock_sleep.assert_called()

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.sleep")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_slow_down(self, mock_time, mock_sleep, mock_post):
        """Test run_auth_flow handles slow_down error"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        slow_down_response = MagicMock()
        slow_down_response.status_code = 400
        slow_down_response.json.return_value = {"error": "slow_down"}

        mock_post.side_effect = [device_response, slow_down_response] + [
            slow_down_response
        ] * 10
        mock_time.side_effect = [0, 1] + [1000] * 20

        result = run_auth_flow(client_id=self.client_id, timeout=10)

        self.assertIsNone(result)
        # Should have increased interval
        mock_sleep.assert_called()

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_access_denied(self, mock_time, mock_post):
        """Test run_auth_flow handles access_denied error"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        denied_response = MagicMock()
        denied_response.status_code = 400
        denied_response.json.return_value = {"error": "access_denied"}

        mock_post.side_effect = [device_response, denied_response]
        mock_time.side_effect = [0, 1, 2]

        result = run_auth_flow(client_id=self.client_id)

        self.assertIsNone(result)

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_expired_token(self, mock_time, mock_post):
        """Test run_auth_flow handles expired_token error"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        expired_response = MagicMock()
        expired_response.status_code = 400
        expired_response.json.return_value = {"error": "expired_token"}

        mock_post.side_effect = [device_response, expired_response]
        mock_time.side_effect = [0, 1, 2]

        result = run_auth_flow(client_id=self.client_id)

        self.assertIsNone(result)

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_unknown_error(self, mock_time, mock_post):
        """Test run_auth_flow handles unknown errors"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        error_response = MagicMock()
        error_response.status_code = 400
        error_response.json.return_value = {"error": "unknown_error"}

        mock_post.side_effect = [device_response, error_response]
        mock_time.side_effect = [0, 1, 2]

        result = run_auth_flow(client_id=self.client_id)

        self.assertIsNone(result)

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.sleep")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_network_error_during_polling(
        self, mock_time, mock_sleep, mock_post
    ):
        """Test run_auth_flow handles network errors during polling"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        mock_post.side_effect = [
            device_response,
            requests.RequestException("Network error"),
        ] + [device_response] * 10
        mock_time.side_effect = [0, 1] + [1000] * 20

        result = run_auth_flow(client_id=self.client_id, timeout=10)

        self.assertIsNone(result)
        mock_sleep.assert_called()

    # Tests for save_credentials_to_db
    @patch("auth.oauth.OAuthCredentials")
    def test_save_credentials_to_db_creates_new(self, mock_oauth_creds):
        """Test saving new credentials to database"""
        mock_row = MagicMock()
        mock_row.id = 1
        mock_oauth_creds.get_or_create.return_value = (mock_row, True)

        result = save_credentials_to_db(
            self.mock_creds,
            client_id=self.client_id,
            client_secret=self.client_secret,
            user_email=self.user_email,
            user_id=self.user_id,
        )

        self.assertEqual(result, 1)
        mock_oauth_creds.get_or_create.assert_called_once()

    @patch("auth.oauth.OAuthCredentials")
    def test_save_credentials_to_db_updates_existing(self, mock_oauth_creds):
        """Test updating existing credentials in database"""
        mock_row = MagicMock()
        mock_row.id = 1
        mock_oauth_creds.get_or_create.return_value = (mock_row, False)

        result = save_credentials_to_db(
            self.mock_creds,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

        self.assertEqual(result, 1)
        mock_row.save.assert_called_once()

    @patch("auth.oauth.OAuthCredentials")
    def test_save_credentials_to_db_handles_no_scopes(self, mock_oauth_creds):
        """Test saving credentials without scopes"""
        mock_row = MagicMock()
        mock_row.id = 1
        self.mock_creds.scopes = None
        mock_oauth_creds.get_or_create.return_value = (mock_row, True)

        result = save_credentials_to_db(self.mock_creds)

        self.assertEqual(result, 1)

    @patch("auth.oauth.OAuthCredentials")
    def test_save_credentials_to_db_handles_scopes_join_exception(
        self, mock_oauth_creds
    ):
        """Test saving credentials when joining scopes fails"""
        mock_row = MagicMock()
        mock_row.id = 1
        # Create a mock that raises when iterated
        mock_scopes = MagicMock()
        mock_scopes.__iter__ = MagicMock(side_effect=Exception("Iterator error"))
        self.mock_creds.scopes = mock_scopes
        mock_oauth_creds.get_or_create.return_value = (mock_row, True)

        result = save_credentials_to_db(self.mock_creds)

        self.assertEqual(result, 1)

    # Tests for load_credentials_from_db
    @patch("auth.oauth.OAuthCredentials")
    def test_load_credentials_from_db_success(self, mock_oauth_creds):
        """Test successfully loading credentials from database"""
        mock_row = MagicMock()
        mock_row.access_token = self.access_token
        mock_row.refresh_token = self.refresh_token
        mock_row.token_uri = TOKEN_URL
        mock_row.client_id = self.client_id
        mock_row.client_secret = self.client_secret
        mock_row.scopes = " ".join(self.scopes)
        mock_row.expiry = datetime.now(timezone.utc) + timedelta(hours=1)

        mock_oauth_creds.select().first.return_value = mock_row

        result = load_credentials_from_db()

        self.assertIsNotNone(result)
        self.assertIsInstance(result, Credentials)
        self.assertEqual(result.token, self.access_token)
        self.assertEqual(result.refresh_token, self.refresh_token)

    @patch("auth.oauth.OAuthCredentials")
    def test_load_credentials_from_db_no_credentials(self, mock_oauth_creds):
        """Test loading credentials when none exist"""
        mock_oauth_creds.select().first.return_value = None

        result = load_credentials_from_db()

        self.assertIsNone(result)

    @patch("auth.oauth.OAuthCredentials")
    def test_load_credentials_from_db_with_user_id(self, mock_oauth_creds):
        """Test loading credentials for specific user"""
        mock_row = MagicMock()
        mock_row.access_token = self.access_token
        mock_row.refresh_token = self.refresh_token
        mock_row.token_uri = TOKEN_URL
        mock_row.client_id = self.client_id
        mock_row.client_secret = self.client_secret
        mock_row.scopes = " ".join(self.scopes)
        mock_row.expiry = datetime.now(timezone.utc) + timedelta(hours=1)

        mock_oauth_creds.get_or_none.return_value = mock_row

        result = load_credentials_from_db(user_id=self.user_id)

        self.assertIsNotNone(result)
        mock_oauth_creds.get_or_none.assert_called_once()

    @patch("auth.oauth.OAuthCredentials")
    def test_load_credentials_from_db_handles_exception(self, mock_oauth_creds):
        """Test loading credentials handles exceptions"""
        mock_oauth_creds.select().first.side_effect = Exception("DB error")

        result = load_credentials_from_db()

        self.assertIsNone(result)

    # Tests for get_credentials
    @patch("auth.oauth.load_credentials_from_db")
    def test_get_credentials_fresh_token(self, mock_load_creds):
        """Test getting fresh credentials that don't need refresh"""
        self.mock_creds.token_state = TokenState.FRESH
        mock_load_creds.return_value = self.mock_creds

        result = get_credentials()

        self.assertEqual(result, self.mock_creds)
        mock_load_creds.assert_called_once()

    @patch("auth.oauth.save_credentials_to_db")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.Request")
    def test_get_credentials_refreshes_stale_token(
        self, mock_request_class, mock_load_creds, mock_save_creds
    ):
        """Test getting credentials refreshes stale token"""
        self.mock_creds.token_state = TokenState.STALE
        self.mock_creds.refresh_token = self.refresh_token
        mock_load_creds.return_value = self.mock_creds

        result = get_credentials()

        self.assertEqual(result, self.mock_creds)
        self.mock_creds.refresh.assert_called_once()
        mock_save_creds.assert_called_once()

    @patch("auth.oauth.save_credentials_to_db")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.Request")
    def test_get_credentials_refreshes_invalid_token(
        self, mock_request_class, mock_load_creds, mock_save_creds
    ):
        """Test getting credentials refreshes invalid token"""
        self.mock_creds.token_state = TokenState.INVALID
        self.mock_creds.refresh_token = self.refresh_token
        mock_load_creds.return_value = self.mock_creds

        result = get_credentials()

        self.assertEqual(result, self.mock_creds)
        self.mock_creds.refresh.assert_called_once()

    @patch("auth.oauth.load_credentials_from_db")
    def test_get_credentials_no_credentials(self, mock_load_creds):
        """Test getting credentials when none exist"""
        mock_load_creds.return_value = None

        result = get_credentials()

        self.assertIsNone(result)

    @patch("auth.oauth.OAuthCredentials")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.Request")
    def test_get_credentials_refresh_fails(
        self, mock_request_class, mock_load_creds, mock_oauth_creds
    ):
        """Test getting credentials when refresh fails"""
        self.mock_creds.token_state = TokenState.STALE
        self.mock_creds.refresh_token = self.refresh_token
        self.mock_creds.refresh.side_effect = Exception("Refresh failed")
        mock_load_creds.return_value = self.mock_creds

        result = get_credentials()

        self.assertIsNone(result)
        mock_oauth_creds.delete().execute.assert_called_once()

    @patch("auth.oauth.OAuthCredentials")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.Request")
    def test_get_credentials_refresh_fails_with_user_id(
        self, mock_request_class, mock_load_creds, mock_oauth_creds
    ):
        """Test getting credentials when refresh fails for specific user"""
        self.mock_creds.token_state = TokenState.STALE
        self.mock_creds.refresh_token = self.refresh_token
        self.mock_creds.refresh.side_effect = Exception("Refresh failed")
        mock_load_creds.return_value = self.mock_creds

        result = get_credentials(user_id=self.user_id)

        self.assertIsNone(result)
        mock_oauth_creds.delete().where().execute.assert_called_once()

    @patch("auth.oauth.load_credentials_from_db")
    def test_get_credentials_no_refresh_token(self, mock_load_creds):
        """Test getting credentials when token is stale but no refresh token"""
        self.mock_creds.token_state = TokenState.STALE
        self.mock_creds.refresh_token = None
        mock_load_creds.return_value = self.mock_creds

        result = get_credentials()

        # Should return credentials without attempting refresh
        self.assertEqual(result, self.mock_creds)

    # Tests for revoke_credentials
    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.OAuthCredentials")
    def test_revoke_credentials_success_with_refresh_token(
        self, mock_oauth_creds, mock_post
    ):
        """Test successfully revoking credentials using refresh token"""
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        result = revoke_credentials(self.mock_creds)

        self.assertTrue(result)
        mock_post.assert_called_once_with(
            REVOKE_URL, params={"token": self.refresh_token}, timeout=10
        )
        mock_oauth_creds.delete().where().execute.assert_called_once()

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.OAuthCredentials")
    def test_revoke_credentials_success_with_access_token(
        self, mock_oauth_creds, mock_post
    ):
        """Test successfully revoking credentials using access token"""
        self.mock_creds.refresh_token = None
        mock_response = MagicMock()
        mock_response.status_code = 204
        mock_post.return_value = mock_response

        result = revoke_credentials(self.mock_creds)

        self.assertTrue(result)
        mock_post.assert_called_once_with(
            REVOKE_URL, params={"token": self.access_token}, timeout=10
        )

    def test_revoke_credentials_no_token(self):
        """Test revoking credentials when no token available"""
        self.mock_creds.refresh_token = None
        self.mock_creds.token = None

        result = revoke_credentials(self.mock_creds)

        self.assertFalse(result)

    @patch("auth.oauth.requests.post")
    def test_revoke_credentials_fails_with_error_status(self, mock_post):
        """Test revoking credentials fails with error status code"""
        mock_response = MagicMock()
        mock_response.status_code = 400
        mock_response.text = "Bad request"
        mock_post.return_value = mock_response

        result = revoke_credentials(self.mock_creds)

        self.assertFalse(result)

    @patch("auth.oauth.requests.post")
    def test_revoke_credentials_network_error(self, mock_post):
        """Test revoking credentials handles network errors"""
        mock_post.side_effect = requests.RequestException("Network error")

        result = revoke_credentials(self.mock_creds)

        self.assertFalse(result)

    @patch("auth.oauth.requests.post")
    @patch("auth.oauth.time.time")
    def test_run_auth_flow_invalid_json_response(self, mock_time, mock_post):
        """Test run_auth_flow handles invalid JSON in error response"""
        device_response = MagicMock()
        device_response.json.return_value = {
            "device_code": "test_device_code",
            "user_code": "TEST-CODE",
            "verification_url": "https://google.com/device",
            "expires_in": 600,
            "interval": 5,
        }
        device_response.raise_for_status = MagicMock()

        error_response = MagicMock()
        error_response.status_code = 400
        error_response.json.side_effect = ValueError("Invalid JSON")

        mock_post.side_effect = [device_response, error_response]
        mock_time.side_effect = [0, 1, 2]

        result = run_auth_flow(client_id=self.client_id)

        self.assertIsNone(result)

    @patch("auth.oauth.OAuthCredentials")
    @patch("auth.oauth.load_credentials_from_db")
    @patch("auth.oauth.Request")
    def test_get_credentials_refresh_fails_delete_exception(
        self, mock_request_class, mock_load_creds, mock_oauth_creds
    ):
        """Test getting credentials when refresh and delete both fail"""
        self.mock_creds.token_state = TokenState.STALE
        self.mock_creds.refresh_token = self.refresh_token
        self.mock_creds.refresh.side_effect = Exception("Refresh failed")
        mock_load_creds.return_value = self.mock_creds

        # Make the delete operation also fail
        mock_oauth_creds.delete().execute.side_effect = Exception("Delete failed")

        result = get_credentials()

        self.assertIsNone(result)

    # Tests for constants
    def test_constants_defined(self):
        """Test that OAuth constants are properly defined"""
        self.assertEqual(DEVICE_CODE_URL, "https://oauth2.googleapis.com/device/code")
        self.assertEqual(TOKEN_URL, "https://oauth2.googleapis.com/token")
        self.assertEqual(REVOKE_URL, "https://oauth2.googleapis.com/revoke")
        self.assertEqual(
            DEFAULT_SCOPES, ["https://www.googleapis.com/auth/youtube.readonly"]
        )



import sys
from unittest import TestCase
from unittest.mock import MagicMock, patch, call

from main import main


class TestMainIntegration(TestCase):
    """Integration tests for the main application entry point"""

    def setUp(self):
        """Set up test fixtures"""
        self.mock_youtube_service = MagicMock()

    def tearDown(self):
        """Clean up after tests"""
        pass

    # Tests for main function - database initialization
    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    @patch("main.YouTubeChannel")
    @patch("main.OAuthCredentials")
    def test_main_creates_youtube_channel_table_if_missing(
        self,
        mock_oauth_creds,
        mock_youtube_channel,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main creates YouTubeChannel table if it doesn't exist"""
        # Setup: YouTubeChannel table missing, OAuthCredentials exists
        mock_database.table_exists.side_effect = (
            lambda table: table == "oauth_credentials"
        )
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify YouTubeChannel table was created
        mock_database.create_tables.assert_called_once_with([mock_youtube_channel])
        mock_database.commit.assert_called()

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    @patch("main.YouTubeChannel")
    @patch("main.OAuthCredentials")
    def test_main_creates_oauth_credentials_table_if_missing(
        self,
        mock_oauth_creds,
        mock_youtube_channel,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main creates OAuthCredentials table if it doesn't exist"""
        # Setup: OAuthCredentials table missing, YouTubeChannel exists
        mock_database.table_exists.side_effect = lambda table: table == "youtubechannel"
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify OAuthCredentials table was created
        assert mock_database.create_tables.call_count == 1
        mock_database.create_tables.assert_called_with([mock_oauth_creds])

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    @patch("main.YouTubeChannel")
    @patch("main.OAuthCredentials")
    def test_main_creates_both_tables_if_missing(
        self,
        mock_oauth_creds,
        mock_youtube_channel,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main creates both tables if neither exist"""
        # Setup: Both tables missing
        mock_database.table_exists.return_value = False
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify both tables were created
        assert mock_database.create_tables.call_count == 2
        mock_database.create_tables.assert_any_call([mock_youtube_channel])
        mock_database.create_tables.assert_any_call([mock_oauth_creds])

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_triggers_force_auth_when_oauth_table_created(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main triggers force_auth when OAuthCredentials table is created"""
        # Setup: OAuthCredentials table missing
        mock_database.table_exists.side_effect = lambda table: table == "youtubechannel"
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify force_auth was called during table creation
        calls = mock_oauth.get_authenticated_youtube_service.call_args_list
        # Should be called twice: once during setup (force_auth=True), once in loop (force_auth=True)
        assert len(calls) >= 2
        assert calls[0] == call(force_auth=True)

    # Tests for main function - main loop
    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_runs_subscription_and_video_checks(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main runs pull_youtube_subscriptions and check_for_new_videos"""
        # Setup: Tables exist
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify subscription pull and video check were called
        mock_pull_subs.assert_called_once_with(self.mock_youtube_service)
        mock_check_videos.assert_called_once_with(self.mock_youtube_service)

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_sleeps_for_calculated_interval(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main sleeps for the calculated interval between checks"""
        # Setup
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        expected_interval = 120
        mock_calc_interval.return_value = expected_interval

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify sleep was called with correct interval
        mock_sleep.assert_called_once_with(expected_interval)

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_calculates_interval_once_at_startup(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that calculate_interval_between_cycles is called once at startup"""
        # Setup
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run twice then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = [None, KeyboardInterrupt()]

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify calculate_interval was called only once (not in the loop)
        mock_calc_interval.assert_called_once()

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_loops_multiple_times(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main continues to loop multiple times"""
        # Setup
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run 3 times then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = [None, None, KeyboardInterrupt()]

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify functions were called 3 times
        assert mock_pull_subs.call_count == 3
        assert mock_check_videos.call_count == 3
        assert mock_sleep.call_count == 3

    # Tests for main function - error handling
    @patch("main.calculate_interval_between_cycles")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_exits_when_no_youtube_service_available(
        self, mock_database, mock_oauth, mock_calc_interval
    ):
        """Test that main exits with code 1 when YouTube service is not available"""
        # Setup: No YouTube service available
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = None
        mock_calc_interval.return_value = 60

        with self.assertRaises(SystemExit) as cm:
            main()

        self.assertEqual(cm.exception.code, 1)

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_gets_fresh_service_each_loop_iteration(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Test that main gets a fresh YouTube service on each loop iteration"""
        # Setup
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Make the loop run 2 times then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = [None, KeyboardInterrupt()]

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify get_authenticated_youtube_service was called twice (once per iteration)
        assert mock_oauth.get_authenticated_youtube_service.call_count == 2

    # Tests for main entry point with command line arguments
    @patch("main.healthcheck")
    @patch("main.load_dotenv")
    @patch("sys.argv", ["main.py", "healthcheck"])
    def test_main_entry_point_runs_healthcheck(
        self, mock_load_dotenv, mock_healthcheck
    ):
        """Test that main entry point runs healthcheck when argv[1] is 'healthcheck'"""
        # Import and execute the if __name__ == "__main__" block
        with patch("main.__name__", "__main__"):
            # Re-execute the module's main block
            exec(
                """
if len(sys.argv) > 1 and sys.argv[1] == "healthcheck":
    healthcheck()
else:
    pass  # Don't run main() in test
""",
                {"sys": sys, "healthcheck": mock_healthcheck, "main": lambda: None},
            )

        mock_healthcheck.assert_called_once()

    @patch("main.main")
    @patch("main.load_dotenv")
    @patch("sys.argv", ["main.py"])
    def test_main_entry_point_runs_main_without_args(self, mock_load_dotenv, mock_main):
        """Test that main entry point runs main() when no arguments provided"""
        # Import and execute the if __name__ == "__main__" block
        with patch("main.__name__", "__main__"):
            exec(
                """
if len(sys.argv) > 1 and sys.argv[1] == "healthcheck":
    pass  # Don't run healthcheck in test
else:
    main()
""",
                {"sys": sys, "healthcheck": lambda: None, "main": mock_main},
            )

        mock_main.assert_called_once()

    @patch("main.main")
    @patch("main.load_dotenv")
    @patch("sys.argv", ["main.py", "other_arg"])
    def test_main_entry_point_runs_main_with_non_healthcheck_arg(
        self, mock_load_dotenv, mock_main
    ):
        """Test that main entry point runs main() when argument is not 'healthcheck'"""
        # Import and execute the if __name__ == "__main__" block
        with patch("main.__name__", "__main__"):
            exec(
                """
if len(sys.argv) > 1 and sys.argv[1] == "healthcheck":
    pass  # Don't run healthcheck in test
else:
    main()
""",
                {"sys": sys, "healthcheck": lambda: None, "main": mock_main},
            )

        mock_main.assert_called_once()

    @patch("main.healthcheck")
    @patch("main.main")
    @patch("main.load_dotenv")
    def test_main_entry_point_loads_dotenv(
        self, mock_load_dotenv, mock_main, mock_healthcheck
    ):
        """Test that main entry point loads .env file"""
        with patch("sys.argv", ["main.py"]):
            with patch("main.__name__", "__main__"):
                exec(
                    """
load_dotenv()
if len(sys.argv) > 1 and sys.argv[1] == "healthcheck":
    healthcheck()
else:
    main()
""",
                    {
                        "sys": sys,
                        "load_dotenv": mock_load_dotenv,
                        "healthcheck": mock_healthcheck,
                        "main": mock_main,
                    },
                )

        mock_load_dotenv.assert_called_once()

    # Integration tests - full flow scenarios
    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_full_flow_with_existing_tables(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Integration test: Full flow when tables already exist"""
        # Setup: Tables exist, service available
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 90

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify full flow
        mock_database.table_exists.assert_any_call("youtubechannel")
        mock_database.table_exists.assert_any_call("oauth_credentials")
        mock_calc_interval.assert_called_once()
        mock_oauth.get_authenticated_youtube_service.assert_called_with(force_auth=True)
        mock_pull_subs.assert_called_once_with(self.mock_youtube_service)
        mock_check_videos.assert_called_once_with(self.mock_youtube_service)
        mock_sleep.assert_called_once_with(90)

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    @patch("main.YouTubeChannel")
    @patch("main.OAuthCredentials")
    def test_main_full_flow_with_table_creation(
        self,
        mock_oauth_creds,
        mock_youtube_channel,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Integration test: Full flow with table creation"""
        # Setup: Tables don't exist
        mock_database.table_exists.return_value = False
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 100

        # Make the loop run once then raise KeyboardInterrupt to exit
        mock_sleep.side_effect = KeyboardInterrupt()

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify full flow including table creation
        assert mock_database.create_tables.call_count == 2
        mock_database.create_tables.assert_any_call([mock_youtube_channel])
        mock_database.create_tables.assert_any_call([mock_oauth_creds])
        assert mock_database.commit.call_count >= 2
        mock_calc_interval.assert_called_once()
        mock_pull_subs.assert_called_once_with(self.mock_youtube_service)
        mock_check_videos.assert_called_once_with(self.mock_youtube_service)
        mock_sleep.assert_called_once_with(100)

    @patch("main.time.sleep")
    @patch("main.calculate_interval_between_cycles")
    @patch("main.check_for_new_videos")
    @patch("main.pull_youtube_subscriptions")
    @patch("main.oauth")
    @patch("main.database")
    def test_main_continuous_operation_simulation(
        self,
        mock_database,
        mock_oauth,
        mock_pull_subs,
        mock_check_videos,
        mock_calc_interval,
        mock_sleep,
    ):
        """Integration test: Simulate continuous operation over multiple cycles"""
        # Setup
        mock_database.table_exists.return_value = True
        mock_oauth.get_authenticated_youtube_service.return_value = (
            self.mock_youtube_service
        )
        mock_calc_interval.return_value = 60

        # Simulate 5 cycles of operation
        mock_sleep.side_effect = [None, None, None, None, KeyboardInterrupt()]

        try:
            main()
        except KeyboardInterrupt:
            pass

        # Verify operations ran 5 times
        assert mock_oauth.get_authenticated_youtube_service.call_count == 5
        assert mock_pull_subs.call_count == 5
        assert mock_check_videos.call_count == 5
        assert mock_sleep.call_count == 5

        # Verify they were called with correct arguments each time
        for call_args in mock_pull_subs.call_args_list:
            self.assertEqual(call_args[0][0], self.mock_youtube_service)
        for call_args in mock_check_videos.call_args_list:
            self.assertEqual(call_args[0][0], self.mock_youtube_service)
        for call_args in mock_sleep.call_args_list:
            self.assertEqual(call_args[0][0], 60)



# create a peewee database instance -- our models will use this database to
# persist information
from peewee import Model, CharField, IntegerField, TextField, DateTimeField, AutoField
from models import database


class BaseModel(Model):
    class Meta:
        database = database


class YouTubeChannel(BaseModel):
    id = CharField(primary_key=True, unique=True, null=False)
    num_videos = IntegerField(null=False)


class OAuthCredentials(BaseModel):
    id = AutoField()
    client_id = CharField(null=True)
    client_secret = TextField(null=True)
    user_id = CharField(null=True)
    user_email = CharField(null=True)
    access_token = TextField(null=True)
    refresh_token = TextField(null=True)
    token_uri = TextField(null=True)
    scopes = TextField(null=True)
    token_type = CharField(null=True)
    expiry = DateTimeField(null=True)
    extra = TextField(null=True)

    class Meta:
        table_name = "oauth_credentials"



from __future__ import annotations

import json
import os
import time
from datetime import datetime, timezone, timedelta
from typing import Optional, Sequence

import requests
from google.auth.credentials import TokenState
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build, Resource

from models.models import OAuthCredentials
from util.logging import logger


DEVICE_CODE_URL = "https://oauth2.googleapis.com/device/code"
TOKEN_URL = "https://oauth2.googleapis.com/token"
REVOKE_URL = "https://oauth2.googleapis.com/revoke"
DEFAULT_SCOPES = ["https://www.googleapis.com/auth/youtube.readonly"]


def get_authenticated_youtube_service(
    force_auth: bool = False,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    scopes: Optional[Sequence[str]] = None,
) -> Resource | None:
    """Return a googleapiclient.discovery.Resource for the YouTube Data API v3.

    - Attempts to load credentials from the DB (and refreshes if needed).
    - If no credentials are present and force_auth is True, runs device auth flow.
    - Returns None when credentials are not available.
    """
    scopes = list(scopes) if scopes else list(DEFAULT_SCOPES)

    if not client_id:
        logger.info("No client_id provided; checking environment variables.")
        client_id = os.environ.get("OAUTH_CLIENT_ID", None)
    if not client_secret:
        logger.info("No client_secret provided; checking environment variables.")
        client_secret = os.environ.get("OAUTH_CLIENT_SECRET", None)

    creds = get_credentials()
    if creds is None and force_auth:
        logger.info("No credentials found; starting device auth flow (console).")
        creds = run_auth_flow(
            client_id=client_id, client_secret=client_secret, scopes=scopes
        )

    if creds is None:
        logger.warning("No valid credentials available for YouTube API.")
        return None

    try:
        service = build("youtube", "v3", credentials=creds)
        return service
    except Exception as exc:
        logger.exception("Failed to build YouTube service: %s", exc)
        return None


def _now_utc() -> datetime:
    return datetime.now(timezone.utc)


def run_auth_flow(
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    scopes: Optional[list[str]] = None,
    timeout: int = 600,
) -> Optional[Credentials]:
    """Run the OAuth2 Device Authorization Flow and persist credentials.

    This logs the verification URL + user code to stdout so an external user
    can complete consent using a browser.
    Returns google.oauth2.credentials.Credentials on success, or None on failure.
    """
    scopes = scopes or DEFAULT_SCOPES

    if client_id is None:
        # Try environment-style client from first DB row or raise
        row = OAuthCredentials.select().first()
        if row and row.client_id:
            client_id = row.client_id

    if client_id is None:
        raise ValueError(
            "client_id must be provided via argument or stored credentials"
        )

    data = {"client_id": client_id, "scope": " ".join(scopes)}

    resp = requests.post(DEVICE_CODE_URL, data=data, timeout=10)
    resp.raise_for_status()
    device = resp.json()

    device_code = device["device_code"]
    user_code = device["user_code"]
    verification_url = device.get("verification_url") or device.get("verification_uri")
    expires_in = int(device.get("expires_in", 600))
    interval = int(device.get("interval", 5))

    logger.info("To authorize, visit:")
    logger.info(verification_url)
    logger.info("and enter the code: %s", user_code)
    logger.info("This code will expire in {} seconds".format(expires_in))

    # Poll token endpoint
    start = time.time()
    while True:
        if time.time() - start > timeout:
            logger.warning("Timeout waiting for user to authorize device code")
            return None

        token_data = {
            "client_id": client_id,
            "device_code": device_code,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        }
        if client_secret:
            token_data["client_secret"] = client_secret

        try:
            t_resp = requests.post(TOKEN_URL, data=token_data, timeout=10)
        except requests.RequestException as exc:
            logger.error("Network error while polling token endpoint: %s", exc)
            time.sleep(interval)
            continue

        if t_resp.status_code == 200:
            tok = t_resp.json()
            creds = Credentials(
                token=tok.get("access_token"),
                refresh_token=tok.get("refresh_token"),
                token_uri=TOKEN_URL,
                client_id=client_id,
                client_secret=client_secret,
                scopes=scopes,
            )
            # expiry may be present
            if "expires_in" in tok:
                creds.expiry = _now_utc() + timedelta(seconds=int(tok["expires_in"]))

            save_credentials_to_db(
                creds, client_id=client_id, client_secret=client_secret
            )
            creds = load_credentials_from_db()
            logger.info("Authorization successful; credentials saved.")
            return creds

        # handle errors
        try:
            err = t_resp.json()
        except ValueError:
            err = {"error": "unknown"}

        error = err.get("error")
        if error == "authorization_pending":
            # user hasn't completed step yet
            time.sleep(interval)
            continue
        if error == "slow_down":
            interval = min(interval + 5, 60)
            time.sleep(interval)
            continue
        if error in ("access_denied", "expired_token"):
            logger.error("Authorization failed: %s", error)
            return None

        # unknown error -> raise
        logger.error("Unexpected token endpoint response: %s", err)
        return None


def save_credentials_to_db(
    creds: Credentials,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    user_email: Optional[str] = None,
    user_id: Optional[str] = None,
) -> int:
    """Persist google.oauth2.credentials.Credentials to the DB using OAuthCredentials model."""
    # join scopes
    scopes = None
    try:
        scopes = " ".join(creds.scopes) if creds.scopes else None
    except Exception:
        scopes = None

    row, created = OAuthCredentials.get_or_create(
        user_id=user_id,
        defaults={
            "client_id": client_id or getattr(creds, "client_id", None),
            "client_secret": client_secret or getattr(creds, "client_secret", None),
            "user_email": user_email,
            "access_token": creds.token,
            "refresh_token": getattr(creds, "refresh_token", None),
            "token_uri": getattr(creds, "token_uri", TOKEN_URL),
            "scopes": scopes,
            "token_type": None,
            "expiry": creds.expiry.astimezone(timezone.utc).replace(tzinfo=None),
            "extra": json.dumps({}),
        },
    )

    if not created:
        # update existing
        row.client_id = client_id or getattr(creds, "client_id", row.client_id)
        row.client_secret = client_secret or getattr(
            creds, "client_secret", row.client_secret
        )
        row.user_email = user_email or row.user_email
        row.access_token = creds.token
        row.refresh_token = getattr(creds, "refresh_token", row.refresh_token)
        row.token_uri = getattr(creds, "token_uri", row.token_uri)
        row.scopes = scopes or row.scopes
        row.expiry = creds.expiry.astimezone(timezone.utc).replace(tzinfo=None)
        row.extra = json.dumps({})
        row.save()

    return row.id


def load_credentials_from_db(user_id: Optional[str] = None) -> Optional[Credentials]:
    try:
        if user_id:
            row = OAuthCredentials.get_or_none(OAuthCredentials.user_id == user_id)
        else:
            row = OAuthCredentials.select().first()
        if not row:
            return None

        creds = Credentials(
            token=row.access_token,
            refresh_token=row.refresh_token,
            token_uri=row.token_uri,
            client_id=row.client_id,
            client_secret=row.client_secret,
            scopes=(row.scopes or "").split(),
        )

        creds.expiry = row.expiry.astimezone(timezone.utc).replace(tzinfo=None)
        return creds
    except Exception as exc:
        logger.error("Error loading credentials from DB: %s", exc)
        return None


def get_credentials(user_id: Optional[str] = None) -> Optional[Credentials]:
    """Return valid credentials, refreshing automatically if needed.

    Returns None if no credentials available or if refresh fails.
    """
    creds = load_credentials_from_db(user_id=user_id)
    if creds is None:
        return None

    # if expired and refresh token present, refresh
    if (
        creds.token_state in (TokenState.STALE, TokenState.INVALID)
        and creds.refresh_token
    ):
        try:
            request = Request()
            creds.refresh(request)
            # persist new tokens
            save_credentials_to_db(
                creds, client_id=creds.client_id, client_secret=creds.client_secret
            )
        except Exception as exc:
            # likely invalid_grant or revoked token
            logger.error("Failed to refresh credentials: %s", exc)
            try:
                # delete the row
                if user_id:
                    OAuthCredentials.delete().where(
                        OAuthCredentials.user_id == user_id
                    ).execute()
                else:
                    OAuthCredentials.delete().execute()
            except Exception:
                pass
            return None

    return creds


def revoke_credentials(creds: Credentials) -> bool:
    token = creds.refresh_token or creds.token
    if not token:
        return False
    try:
        resp = requests.post(REVOKE_URL, params={"token": token}, timeout=10)
        if resp.status_code in (200, 204):
            # remove from DB
            OAuthCredentials.delete().where(
                OAuthCredentials.client_id == creds.client_id
            ).execute()
            return True
        logger.error("Failed to revoke token: %s %s", resp.status_code, resp.text)
        return False
    except requests.RequestException as exc:
        logger.error("Network error revoking token: %s", exc)
        return False



from pathlib import Path
import os
from peewee import SqliteDatabase

# Allow overriding path via env var; default to data/bsn.db
_db_path = Path(os.getenv("DATA_DIR", "../data")) / "bsn.db"
_db_path.parent.mkdir(parents=True, exist_ok=True)

database = SqliteDatabase(str(_db_path))



from unittest import TestCase
from unittest.mock import MagicMock, patch

from notifications.notifications import send_youtube_channels_notifications


class TestNotifications(TestCase):
    def setUp(self):
        """Set up test fixtures"""
        self.sample_channel = {
            "id": "UC1234567890",
            "snippet": {"title": "Test Channel"},
            "statistics": {"videoCount": "100"},
        }
        self.sample_channel_2 = {
            "id": "UC0987654321",
            "snippet": {"title": "Another Channel"},
            "statistics": {"videoCount": "200"},
        }
        self.sample_video = {
            "snippet": {
                "title": "Amazing Test Video",
                "resourceId": {"videoId": "dQw4w9WgXcQ"},
            }
        }

    def tearDown(self):
        """Clean up after tests"""
        pass

    # Tests for send_youtube_channels_notifications
    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_single_channel_with_video(self, mock_apprise_class):
        """Test sending notification for a single channel with video details"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        send_youtube_channels_notifications([self.sample_channel], self.sample_video)

        # Verify Apprise was instantiated
        mock_apprise_class.assert_called_once()

        # Verify the apprise URL was added
        mock_apprise_instance.add.assert_called_once_with("test://localhost")

        # Verify notify was called with correct title and body
        expected_title = "Test Channel has uploaded a new video to YouTube!"
        expected_body = (
            "Amazing Test Video\nhttps://www.youtube.com/watch?v=dQw4w9WgXcQ"
        )
        mock_apprise_instance.notify.assert_called_once_with(
            title=expected_title, body=expected_body
        )

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_multiple_channels(self, mock_apprise_class):
        """Test sending notification for multiple channels"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        channels = [self.sample_channel, self.sample_channel_2]
        send_youtube_channels_notifications(channels, None)

        # Verify Apprise was instantiated
        mock_apprise_class.assert_called_once()

        # Verify the apprise URL was added
        mock_apprise_instance.add.assert_called_once_with("test://localhost")

        # Verify notify was called with correct title and body
        expected_title = (
            "Test Channel,Another Channel have uploaded new videos to YouTube!"
        )
        expected_body = (
            "Check them out here: https://www.youtube.com/feed/subscriptions"
        )
        mock_apprise_instance.notify.assert_called_once_with(
            title=expected_title, body=expected_body
        )

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_single_channel_without_video(self, mock_apprise_class):
        """Test sending notification for a single channel without video details (should not send)"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        send_youtube_channels_notifications([self.sample_channel], None)

        # Verify Apprise was instantiated
        mock_apprise_class.assert_called_once()

        # Verify the apprise URL was added
        mock_apprise_instance.add.assert_called_once_with("test://localhost")

        # Verify notify was NOT called (returns early with warning)
        mock_apprise_instance.notify.assert_not_called()

    @patch("notifications.notifications.apprise.Apprise")
    @patch(
        "notifications.notifications.apprise_urls",
        ["test://localhost", "test://example.com"],
    )
    def test_send_notifications_multiple_apprise_urls(self, mock_apprise_class):
        """Test that multiple apprise URLs are added"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        send_youtube_channels_notifications([self.sample_channel], self.sample_video)

        # Verify both apprise URLs were added
        assert mock_apprise_instance.add.call_count == 2
        mock_apprise_instance.add.assert_any_call("test://localhost")
        mock_apprise_instance.add.assert_any_call("test://example.com")

        # Verify notify was still called once
        mock_apprise_instance.notify.assert_called_once()

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_three_channels(self, mock_apprise_class):
        """Test sending notification for three channels"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        channel_3 = {
            "id": "UC1111111111",
            "snippet": {"title": "Third Channel"},
            "statistics": {"videoCount": "300"},
        }
        channels = [self.sample_channel, self.sample_channel_2, channel_3]
        send_youtube_channels_notifications(channels, None)

        # Verify notify was called with all three channel names
        expected_title = "Test Channel,Another Channel,Third Channel have uploaded new videos to YouTube!"
        expected_body = (
            "Check them out here: https://www.youtube.com/feed/subscriptions"
        )
        mock_apprise_instance.notify.assert_called_once_with(
            title=expected_title, body=expected_body
        )

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_video_with_special_characters(self, mock_apprise_class):
        """Test sending notification with video title containing special characters"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        special_video = {
            "snippet": {
                "title": "Test Video: Amazing & Cool! (2026)",
                "resourceId": {"videoId": "abc123"},
            }
        }

        send_youtube_channels_notifications([self.sample_channel], special_video)

        # Verify notify was called with the special characters preserved
        expected_body = (
            "Test Video: Amazing & Cool! (2026)\nhttps://www.youtube.com/watch?v=abc123"
        )
        mock_apprise_instance.notify.assert_called_once()
        call_args = mock_apprise_instance.notify.call_args
        assert call_args[1]["body"] == expected_body

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_channel_with_special_characters(
        self, mock_apprise_class
    ):
        """Test sending notification with channel name containing special characters"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        special_channel = {
            "id": "UC1234567890",
            "snippet": {"title": "Test & Demo Channel™"},
            "statistics": {"videoCount": "100"},
        }

        send_youtube_channels_notifications([special_channel], self.sample_video)

        # Verify notify was called with the special characters preserved in title
        expected_title = "Test & Demo Channel™ has uploaded a new video to YouTube!"
        mock_apprise_instance.notify.assert_called_once()
        call_args = mock_apprise_instance.notify.call_args
        assert call_args[1]["title"] == expected_title

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", [])
    def test_send_notifications_empty_apprise_urls(self, mock_apprise_class):
        """Test behavior when no apprise URLs are configured"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        send_youtube_channels_notifications([self.sample_channel], self.sample_video)

        # Verify Apprise was instantiated
        mock_apprise_class.assert_called_once()

        # Verify add was never called (no URLs)
        mock_apprise_instance.add.assert_not_called()

        # Verify notify was still called (Apprise handles no destinations gracefully)
        mock_apprise_instance.notify.assert_called_once()

    @patch("notifications.notifications.logger")
    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_logs_warning_for_no_video(
        self, mock_apprise_class, mock_logger
    ):
        """Test that a warning is logged when single channel has no video"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        send_youtube_channels_notifications([self.sample_channel], None)

        # Verify warning was logged
        mock_logger.warning.assert_called_once_with("No video found")

    @patch("notifications.notifications.logger")
    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_logs_info_for_multiple_channels(
        self, mock_apprise_class, mock_logger
    ):
        """Test that info is logged when sending notifications for multiple channels"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        channels = [self.sample_channel, self.sample_channel_2]
        send_youtube_channels_notifications(channels, None)

        # Verify info was logged
        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]
        assert "Sending notifications for" in call_args
        assert "channels" in call_args

    @patch("notifications.notifications.logger")
    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_logs_info_for_single_channel_with_video(
        self, mock_apprise_class, mock_logger
    ):
        """Test that info is logged when sending notifications for single channel with video"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        send_youtube_channels_notifications([self.sample_channel], self.sample_video)

        # Verify info was logged
        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]
        assert "Sending notifications for" in call_args
        assert "channels" in call_args

    @patch("notifications.notifications.apprise.Apprise")
    @patch("notifications.notifications.apprise_urls", ["test://localhost"])
    def test_send_notifications_empty_channels_list(self, mock_apprise_class):
        """Test handling of empty channels list"""
        mock_apprise_instance = MagicMock()
        mock_apprise_class.return_value = mock_apprise_instance

        # This should raise an IndexError or be handled
        # Based on the code, it will try to access channels[0] when len != > 1
        # and video is provided, which will fail
        # Let's test the case where no video is provided with empty list
        with self.assertRaises(IndexError):
            send_youtube_channels_notifications([], self.sample_video)
